<#
    .SYNOPSIS
        Helper functions used by various scripts.
    .DESCRIPTION
        These are functions that have been split out for easier maintenance across multiple modules/scripts.
        Ensure any new functions are generic enough to be consumed by multiple Puppet modules if needed.
        Not all Puppet modules will utilize all functions.
        Functions contained within this file should be named in the format of "<verb>-Helper<descriptor>"
    .NOTES
        Origin files may be found here: https://gitlab.com/rfbennett/ps1_extenders
#>

[string]$script:HelperVersion = '2024.12.17'
[string]$script:HelperOutputFile = ''
[string]$script:HelperOutputLead = ''
[boolean]$script:HelperOutputEnabled = $true

[Flags()] enum ServiceAccess {
    QueryConfig = 0x00000001
    ChangeConfig = 0x00000002
    QueryStatus = 0x00000004
    EnumerateDependents = 0x00000008
    Start = 0x00000010
    Stop = 0x00000020
    PauseContinue = 0x00000040
    Interrogate = 0x00000080
    UserDefinedControl = 0x00000100
    Delete = 0x00010000
    ReadControl = 0x00020000
    WriteDac = 0x00040000
    WriteOwner = 0x00080000
    AllAccess = 0x000F01FF
    AccessSystemSecurity = 0x01000000
}

Function Write-HelperOutput {
    [CmdletBinding()]
    Param
    (
        [Parameter(Mandatory = $false)][AllowEmptyString()][string]$Message = '',
        [Parameter(Mandatory = $false)][AllowEmptyString()][string]$Lead = $script:HelperOutputLead,
        [Parameter(Mandatory = $false)][ValidateSet('Console', 'Debug', 'Warning', 'Error', 'Info')][string]$Type = 'Debug',
        [Parameter(Mandatory = $false)][ValidateSet('Black', 'DarkBlue', 'DarkGreen', 'DarkCyan', 'DarkRed', 'DarkMagenta', 'DarkYellow', 'Gray', 'DarkGray', 'Blue', 'Green', 'Cyan', 'Red', 'Magenta', 'Yellow', 'White')][string]$ForegroundColor = 'DarkGray'
    )
    if ("$Message" -ne '') {
        $Message = $Message.Trim()
        if (("$Message" -match 'execution complete$') -or ("$Message" -match 'executing...$') -or ("$Message" -match '^isAlreadySet:')) {
            if ($Lead.Length -gt 1) {
                $Lead = $Lead.Substring(2)
            }
        }
    }
    if ($script:HelperOutputEnabled) {
        if (($Message -ne '') -and (($Type -eq 'Console') -or ($Type -eq 'Error'))) {
            Write-Host "$Message" -ForegroundColor "$ForegroundColor"
        }
    }
    [string]$logType = 'DBG'
    switch ($Type) {
        'Console' { $logType = 'CNS'; break }
        'Warning' { $logType = 'WRN'; break }
        'Error' { $logType = 'ERR'; break }
        'Info' { $logType = 'INF'; break }
    }
    if ('' -ne $script:HelperOutputFile) {
        [boolean]$isExist = Test-Path -Path "$($script:HelperOutputFile)" -ErrorAction 'SilentlyContinue'
        if ($isExist) {
            [string]$user = [System.Security.Principal.WindowsIdentity]::GetCurrent().Name
            [string]$nowDate = Get-Date -format "yyyy-MM-dd, HH:mm:ss"
            "$nowDate, $user, $logType, $($Lead)$Message" | Add-Content -Path "$($script:HelperOutputFile)"
        }
        else {
            Write-Warning "Unable to locate the required '$($script:HelperOutputFile)' log file"
        }
    }
}

Function Initialize-HelperOutput {
    [CmdletBinding()]
    Param (
        [Parameter(Mandatory = $false)][string]$Identifier = "puppet_ps_helper",
        [Parameter(Mandatory = $false)][string]$LogFolder = "$($env:WinDir)\Temp",
        [Parameter(Mandatory = $false)][boolean]$LogOutput = $true,
        [Parameter(Mandatory = $false)][Int64]$MaxLogFileSize = 5 #MBs
    )
    if ($LogOutput) {
        [boolean]$isExist = Test-Path -Path "$LogFolder"  -PathType 'Container' -ErrorAction 'SilentlyContinue'
        if (!$isExist) {
            Write-Host "Creating the '$LogFolder' folder" -ForegroundColor 'DarkGray'
            $null = New-Item -ItemType Directory -Path "$LogFolder"
        }
    }
    [string]$fileName = "$($Identifier)_debug.log"
    $script:HelperOutputFile = "$LogFolder\$fileName"
    $script:HelperOutputEnabled = $LogOutput
    [string]$debugEnvironmentVariable = [System.Environment]::GetEnvironmentVariable('PUPPET_PS_HELPER_DEBUG', [System.EnvironmentVariableTarget]::Machine)
    if ('True' -ne $debugEnvironmentVariable) {
        Write-Host "To enable debug mode, create a SYSTEM environment variable named 'PUPPET_PS_HELPER_DEBUG' and set the value to 'True" -ForegroundColor 'DarkGray'
        if ($script:HelperOutputEnabled -ne $true) {
            return
        }
    }
    else {
        Write-Host "Debug mode enabled for PS1 - you may review the contents of '$script:HelperOutputFile'" -ForegroundColor 'DarkGray'
    }
    [string]$nowDate = Get-Date -format "yyyy-MM-dd, HH:mm:ss"
    [string]$user = [System.Security.Principal.WindowsIdentity]::GetCurrent().Name
    [boolean]$isExist = Test-Path -Path "$script:HelperOutputFile" -ErrorAction 'SilentlyContinue'
    if ($isExist) {
        [int]$fileSize = [System.Math]::Round(((($script:HelperOutputFile).length) / 1MB), 2)
        if ($fileSize -gt $MaxLogFileSize) {
            Write-Host 'Removing ovesized log file' -ForegroundColor 'DarkGray'
            Remove-Item -Path "$script:HelperOutputFile" -Force -ErrorAction 'SilentlyContinue'
            $isExist = Test-Path -Path "$script:HelperOutputFile" -ErrorAction 'SilentlyContinue'
            if ($isExist) {
                Write-Warning "Failed to remove '$script:HelperOutputFile'"
            }
            $null = New-Item -Path "$LogFolder" -Name "$fileName" -ItemType "File" -Value "$nowDate, $user, DBG, Initialize $Identifier log file`r"
        }
    }
    else {
        $null = New-Item -Path "$LogFolder" -Name "$fileName" -ItemType "File" -Value "$nowDate, $user, DBG, Initialize $Identifier log file`r"
    }
    Write-HelperOutput -Message "Helpers Version: $($script:HelperVersion)" -Lead ''
    Write-HelperOutput -Message "Helpers PSScriptRoot: $($PSScriptRoot)" -Lead ''
    return
}

Function Set-HelperExtension {
    [OutputType([string])]
    [CmdletBinding()]
    Param
    (
        [Parameter(Mandatory = $true)][string]$FileName,
        [Parameter(Mandatory = $true)][string]$Extension,
        [Parameter(Mandatory = $false)][boolean]$IsExitOnFailure = $true
    )
    [string]$functionName = "{0}" -f $MyInvocation.MyCommand
    Write-HelperOutput -Message "$functionName executing..."
    Write-HelperOutput -Message "FileName: $FileName"
    Write-HelperOutput -Message "Extension: $Extension"
    Write-HelperOutput -Message "IsExitOnFailure: $IsExitOnFailure"
    try {
        [string]$newFileName = $FileName
        if ($FileName -gt 4) {
            [string]$extensionCheck = $FileName.Substring($FileName.Length - 4)
            if ($extensionCheck.ToLower() -ne "$Extension") {
                Write-HelperOutput -Message "Adding '$Extension' file extension to '$($FileName)'"
                $newFileName = "$($FileName)$($Extension)"
            }
            else {
                Write-HelperOutput -Message 'File extension already set'
            }
        }
        else {
            Write-HelperOutput -Message "Adding missing '$Extension' to the '$FileName' file name"
            $newFileName = "$($FileName)$($Extension)"
        }
    }
    catch {
        Write-HelperOutput -Message "$($functionName): $_" -Type 'Console' -ForegroundColor 'Red'
        if ($IsExitOnFailure) {
            Invoke-HelperExit -Message "An error occured verifying file extension" -ExitCode 1 -Type 'failure'
        }
        else {
            Write-HelperOutput -Message "An unexpected error occured verifying file extension" -ForegroundColor 'Red'
        }
    }
    Write-HelperOutput -Message "$functionName execution complete"
    return $newFileName
}

Function Remove-HelperItem {
    [OutputType([boolean])]
    [CmdletBinding()]
    Param
    (
        [Parameter(Mandatory = $true)][string]$Item,
        [Parameter(Mandatory = $false)][boolean]$IsExitOnFailure = $false
    )
    [string]$functionName = "{0}" -f $MyInvocation.MyCommand
    Write-HelperOutput -Message "$functionName executing..."
    Write-HelperOutput -Message "Item: $Item"
    Write-HelperOutput -Message "IsExitOnFailure: $IsExitOnFailure"
    [boolean]$isRemoved = $false
    try {
        [boolean]$isExist = Test-Path -Path "$Item" -ErrorAction 'SilentlyContinue'
        if ($isExist) {
            Write-HelperOutput -Message 'Removing the existing item'
            Remove-Item -Path "$Item" -Force
            $isExist = Test-Path -Path "$Item" -ErrorAction 'SilentlyContinue'
            if ($isExist) {
                if ($IsExitOnFailure) {
                    Invoke-HelperExit -Message "Unable to remove existing item" -ExitCode 1 -Type 'failure'
                }
                else {
                    Write-HelperOutput -Message "Unable to remove existing item" -Type 'Error'
                }
            }
            else {
                $isRemoved = $true
                Write-HelperOutput -Message 'Successfully removed the item' -ForegroundColor 'Green'
            }
        }
        else {
            Write-HelperOutput -Message 'No existing item found'
        }
    }
    catch {
        Write-HelperOutput -Message "$($functionName): $_" -Type 'Console' -ForegroundColor 'Red'
        if ($IsExitOnFailure) {
            Invoke-HelperExit -Message "An unexpected error occured removing the '$Item' item" -Type 'failure'
        }
        else {
            Write-HelperOutput -Message "An unexpected error occured removing the '$Item' item" -Type 'Error'
        }
    }
    Write-HelperOutput -Message "$functionName execution complete"
    return $isRemoved
}

Function Copy-HelperItem {
    [OutputType([Boolean])]
    [CmdletBinding()]
    Param(
        [Parameter(Mandatory = $true)][string]$SourceItem,
        [Parameter(Mandatory = $true)][string]$Destination,
        [Parameter(Mandatory = $false)][string]$DestinationItemName = '',
        [Parameter(Mandatory = $false)][boolean]$BackupExistingDestinationItem = $false
    )
    [string]$functionName = "{0}" -f $MyInvocation.MyCommand
    Write-HelperOutput -Message "$functionName executing..."
    Write-HelperOutput -Message "Destination: $Destination"
    Write-HelperOutput -Message "SourceItem: $SourceItem"
    Write-HelperOutput -Message "DestinationItemName: $DestinationItemName"
    Write-HelperOutput -Message "BackupExistingDestinationItem: $BackupExistingDestinationItem"
    [boolean]$isSuccess = $false
    [boolean]$isExist = Test-Path -Path "$SourceItem" -ErrorAction 'SilentlyContinue'
    if (!$isExist) {
        Write-HelperOutput -Message "Unable to locate the '$SourceItem' file" -Type 'Warning'
        return $isSuccess
    }
    [string]$sourceItemName = Split-Path -Path "$SourceItem" -Leaf -Resolve -ErrorAction 'SilentlyContinue'
    Write-HelperOutput -Message "sourceItemName: $sourceItemName"
    [string]$destinationItem = "$Destination\$sourceItemName"
    if ('' -ne "$DestinationItemName") {
        $destinationItem = "$Destination\$DestinationItemName"
    }
    Write-HelperOutput -Message "destinationItem: $destinationItem"
    if ($BackupExistingDestinationItem) {
        try {
            $isExist = Test-Path -Path "$destinationItem" -ErrorAction 'SilentlyContinue'
            if ($isExist) {
                [string]$timestamp = Get-Date -format "MM-dd-yyyy.hh-mm-ss"
                Write-HelperOutput -Message "Backing up '$destinationItem' as '$destinationItem.$timestamp.bak'"
                Copy-Item -Path "$destinationItem" -Destination "$destinationItem.$timestamp.bak" -Force -ErrorAction 'SilentlyContinue'
                $isExist = Test-Path -Path "$destinationItem.$timestamp.bak" -ErrorAction 'SilentlyContinue'
                if ($isExist) {
                    Write-HelperOutput -Message 'Backup process completed successfully'
                }
                else {
                    Invoke-HelperExit -Message "Failed to back up '$destinationItem'"
                }
            }
            else {
                Write-HelperOutput -Message "No existing '$destinationItem' item found to back up"
            }
        }
        catch {
            Write-HelperOutput -Message "$($functionName): $_" -Type 'Console' -ForegroundColor 'Red'
            Invoke-HelperExit -Message "An unexpected error occured backing up the existing '$destinationItem' item"
        }
    }
    try {
        [string]$timestamp = Get-Date -format "MM-dd-yyyy.hh-mm-ss"
        Write-HelperOutput -Message "Copying '$SourceItem' as '$destinationItem'"
        [System.IO.FileInfo]$copiedItem = Copy-Item -Path "$SourceItem" -Destination "$destinationItem" -Force -PassThru -ErrorAction 'SilentlyContinue'
        if ($null -eq $copiedItem) {
            Invoke-HelperExit -Message "Failed to copy '$SourceItem'"
        }
        else {
            Write-HelperOutput -Message 'Copy process completed successfully'
            $isSuccess = $true
        }
    }
    catch {
        Write-HelperOutput -Message "$($functionName): $_" -Type 'Console' -ForegroundColor 'Red'
        Invoke-HelperExit -Message "An unexpected error occured copying the '$SourceItem' item"
    }
    Write-HelperOutput "$functionName execution complete"
    return $isSuccess
}

Function Set-HelperFolder {
    [OutputType([boolean])]
    [CmdletBinding()]
    Param
    (
        [Parameter(Mandatory = $true)][string]$Folder,
        [Parameter(Mandatory = $false)][boolean]$IsExitOnFailure = $false
    )
    [string]$functionName = "{0}" -f $MyInvocation.MyCommand
    Write-HelperOutput -Message "$functionName executing..."
    Write-HelperOutput -Message "Folder: $Folder"
    Write-HelperOutput -Message "IsExitOnFailure: $IsExitOnFailure"
    [boolean]$isCreated = $false
    try {
        [boolean]$isExist = Test-Path -Path "$Folder" -ErrorAction 'SilentlyContinue'
        if ($isExist) {
            $isCreated = $true
            Write-HelperOutput -Message 'Found existing folder'
        }
        else {
            Write-HelperOutput -Message 'Creating folder structure'
            [void](New-Item -ItemType Directory -Path "$Folder")
            $isExist = Test-Path -Path "$Folder" -ErrorAction 'SilentlyContinue'
            if ($isExist) {
                $isCreated = $true
                Write-HelperOutput -Message 'Located newly created folder'
            }
            else {
                if ($IsExitOnFailure) {
                    Invoke-HelperExit -Message "Unable to locate the folder" -ExitCode 1 -Type 'failure'
                }
                else {
                    Write-HelperOutput -Message "Unable to locate the folder" -Type 'Error'
                }
            }
        }
    }
    catch {
        Write-HelperOutput -Message "$($functionName): $_" -Type 'Console' -ForegroundColor 'Red'
        if ($IsExitOnFailure) {
            Invoke-HelperExit -Message "An unexpected error occured creating the '$Folder' folder" -Type 'failure'
        }
        else {
            Write-HelperOutput -Message "An unexpected error occured creating the '$Folder' folder" -Type 'Error'
        }
    }
    Write-HelperOutput -Message "$functionName execution complete"
    return $isCreated
}

Function Test-HelperCompare {
    [OutputType([boolean])]
    [CmdletBinding()]
    Param
    (
        [Parameter(Mandatory = $true)]$Value,
        [Parameter(Mandatory = $true)]$OtherValue,
        [Parameter(Mandatory = $false)][string]$Identifier = '',
        [Parameter(Mandatory = $false)][boolean]$IsPriorValid = $true,
        [Parameter(Mandatory = $false)][boolean]$LogOutput = $false
    )
    $script:HelperOutputLead = '    '
    [boolean]$isSame = $true
    if ($LogOutput) {
        if ($Identifier -eq '') {
            Write-HelperOutput -Message "'$($Value.ToString())' vs '$($OtherValue.ToString())'"
        }
        else {
            Write-HelperOutput -Message "$($Identifier): '$($Value.ToString())' vs '$($OtherValue.ToString())'"
        }
    }
    if ($IsPriorValid) {
        if ($Value -ne $OtherValue) {
            if ($Identifier -eq '') {
                Write-HelperOutput -Message "The value of '$Value' does not match '$OtherValue'" -ForegroundColor 'Magenta' -Type 'Console'
            }
            else {
                Write-HelperOutput -Message "The value of '$Value' does not match '$OtherValue' ($Identifier)" -ForegroundColor 'Magenta' -Type 'Console'
            }
            $isSame = $false
        }
    }
    else {
        # No need to compare this one, because a prior comparison already failed
        $isSame = $false
    }
    $script:HelperOutputLead = '  '
    return $isSame
}

Function Get-HelperObjectsFromPuppetString {
    [OutputType([Collections.Generic.List[Object]])]
    [CmdletBinding()]
    Param
    (
        [Parameter(Mandatory = $true)][string]$StringToConvert,
        [Parameter(Mandatory = $true)][string[]]$PropertyNames,
        [Parameter(Mandatory = $false)][boolean]$ReplaceDoubleBackslashedValues = $true
    )
    [string]$functionName = "{0}" -f $MyInvocation.MyCommand
    $script:HelperOutputLead = '    '
    Write-HelperOutput -Message "$functionName executing..."
    Write-HelperOutput -Message "PropertyNames: $PropertyNames"
    Write-HelperOutput -Message "Number of property names: $($PropertyNames.Count)"
    Write-HelperOutput -Message "ReplaceDoubleBackslashedValues: $ReplaceDoubleBackslashedValues"
    Write-HelperOutput -Message "StringToConvert: $StringToConvert"
    $StringToConvert = $StringToConvert.Trim()
    [Collections.Generic.List[Object]]$hashObjects = $null
    [boolean]$isValid = $false
    [int]$hashCount = 0
    [string]$hashString = ''
    [string]$hashElement = ''
    [string]$propertyValue = ''
    [string[]]$hashArray = @()
    [string[]]$valueArray = @()
    [PSCustomObject]$hashStringObject = $null
    [PSCustomObject]$hashObject = $null
    [int]$propertyCount = 0
    [string[]]$objectPropertyNames = @()
    foreach ($propertyName in $PropertyNames) {
        $propertyName = $propertyName.Replace('"', '')
        $propertyName = $propertyName.Replace("'", '')
        $propertyName = $propertyName.Trim()
        $objectPropertyNames += $propertyName
    }
    if (($StringToConvert.StartsWith('[')) -or ($StringToConvert.EndsWith(']'))) {
        # Strip off leading and trailing brackets (ie. an array of hashes has been supplied)
        if ($StringToConvert.StartsWith('[')) {
            $StringToConvert = $StringToConvert.Substring(1, $StringToConvert.length - 1)
        }
        if ($StringToConvert.EndsWith(']')) {
            $StringToConvert = $StringToConvert.Substring(0, $StringToConvert.length - 1)
        }
    }
    else {
        if (($StringToConvert.StartsWith('{')) -or ($StringToConvert.EndsWith('}'))) {
            # Strip off leading and trailing curly brackets (ie. a hash of hashes has been supplied)
            if ($StringToConvert.StartsWith('{')) {
                $StringToConvert = $StringToConvert.Substring(1, $StringToConvert.length - 1)
            }
            if ($StringToConvert.EndsWith('}')) {
                $StringToConvert = $StringToConvert.Substring(0, $StringToConvert.length - 1)
            }
        }
    }
    $hashStringObject = ConvertFrom-String -InputObject $StringToConvert -Delimiter '},'
    if ($null -eq $hashStringObject) {
        $hashStringObject = ConvertFrom-String -InputObject $StringToConvert -Delimiter '}'
    }
    if ($null -eq $hashStringObject) {
        $hashStringObject = ConvertFrom-String -InputObject "{$($StringToConvert)}" -Delimiter '}'
    }
    $hashStringObject.PSObject.Properties | ForEach-Object {
        $hashCount = $hashCount + 1
        $hashObject = New-Object PSObject
        foreach ($objectPropertyName in $objectPropertyNames) {
            $hashObject | Add-Member -MemberType "NoteProperty" -Name "$objectPropertyName" -Value $null
        }
        if (($null -ne $_.Value) -and ('' -ne $_.Value)) {
            $hashArray = $_.Value -Split '{'
            $hashString = $hashArray[1]
            if ('' -ne $hashString) {
                # Strip off any leading or trailing curly brackets
                if ($hashString.StartsWith('{')) {
                    $hashString = $hashString.Substring(1, $hashString.length - 1)
                }
                if ($hashString.EndsWith('}')) {
                    $hashString = $hashString.Substring(0, $hashString.length - 1)
                }
                Write-HelperOutput -Message "hashString $($hashCount): $hashString"
                $propertyCount = 0
                foreach ($objectPropertyName in $objectPropertyNames) {
                    $isValid = $false
                    $propertyValue = $null
                    $propertyCount = $propertyCount + 1
                    $hashArray = $hashString -Split """$($objectPropertyName)""=>"
                    if ($hashArray.Count -le 1) {
                        # Try adding a space this time
                        $hashArray = $hashString -Split """$($objectPropertyName)"" =>"
                        if ($hashArray.Count -le 1) {
                            # Try no quotes this time
                            $hashArray = $hashString -Split "$($objectPropertyName)=>"
                            if ($hashArray.Count -le 1) {
                                # Try no quotes and add a space
                                $hashArray = $hashString -Split "$($objectPropertyName) =>"
                            }
                        }
                    }
                    $hashElement = $hashArray[1]
                    $hashElement = $hashElement.Trim()
                    if ('' -ne $hashElement) {
                        Write-HelperOutput -Message "hashElement $($propertyCount): $hashElement"
                        $script:HelperOutputLead = '      '
                        if ($hashElement.StartsWith('[')) {
                            $valueArray = $hashElement -Split ']'
                            $propertyValue = $valueArray[0]
                            $propertyValue = $propertyValue.Trim()
                            Write-HelperOutput -Message "Pre cleanup Array data: $propertyValue"
                            if ($propertyValue.StartsWith('[')) {
                                # Remove the leading bracket (the trailing one for the element already got dropped)
                                $propertyValue = $propertyValue.Substring(1, $propertyValue.length - 1)
                            }
                        }
                        else {
                            $valueArray = $hashElement -Split ','
                            $propertyValue = $valueArray[0]
                            $propertyValue = $propertyValue.Trim()
                            Write-HelperOutput -Message "Pre cleanup String or Integer data: $propertyValue"
                            [int]$characters = 0
                            [string]$newCharacter = ''
                            if ($propertyValue.StartsWith('"')) {
                                $characters = 1
                            }
                            if ($propertyValue.StartsWith('"\')) {
                                $characters = 2
                            }
                            if ($propertyValue.Length -gt $characters) {
                                $propertyValue = $propertyValue.Substring($characters, $propertyValue.length - $characters)
                            }
                            $characters = 0
                            if ($propertyValue.EndsWith('"')) {
                                $characters = 1
                            }
                            if ($propertyValue.EndsWith('\""')) {
                                $characters = 3
                                $newCharacter = '"'
                            }
                            if ($propertyValue.Length -gt $characters) {
                                $propertyValue = $propertyValue.Substring(0, $propertyValue.length - $characters)
                                $propertyValue = $propertyValue + $newCharacter
                            }
                        }
                        if ($ReplaceDoubleBackslashedValues) {
                            $propertyValue = $propertyValue.Replace('\\', '\')
                            $propertyValue = $propertyValue.Replace('\"', '"')
                        }
                        Write-HelperOutput -Message "objectPropertyName: $objectPropertyName"
                        Write-HelperOutput -Message "propertyValue: $propertyValue"
                        $script:HelperOutputLead = '    '
                        $hashObject."$objectPropertyName" = $propertyValue
                        $isValid = $true
                    }
                }
                if ($isValid) {
                    $hashObjects += $hashObject
                }
            }
        }
    }
    Write-HelperOutput -Message "$functionName execution complete"
    $script:HelperOutputLead = '  '
    return $hashObjects
}

Function Invoke-HelperExit {
    [CmdletBinding()]
    Param
    (
        [Parameter(Mandatory = $false)][string]$Message = '',
        [Parameter(Mandatory = $false)][int]$ExitCode = -1,
        [Parameter(Mandatory = $false)][ValidateSet('success', 'failure')][string]$Type = 'failure'
    )
    if ('' -ne $Message) {
        switch ($Type) {
            'success' { Write-HelperOutput -Message "$Message" -ForegroundColor 'Green' -Type 'Console' -Lead '' }
            'failure' { Write-HelperOutput -Message "$Message" -ForegroundColor 'Red' -Type 'Error' -Lead '' }
        }
    }
    Write-HelperOutput -Message "=================================================" -Lead ''
    Exit $ExitCode
}

Function Test-HelperAdmin {
    [OutputType([boolean])]
    [CmdletBinding()]
    Param (
        [Parameter(Mandatory = $false)][boolean]$IsExitOnFailure = $true
    )
    [string]$functionName = "{0}" -f $MyInvocation.MyCommand
    $script:HelperOutputLead = '    '
    Write-HelperOutput -Message "$functionName executing..."
    Write-HelperOutput -Message "IsExitOnFailure: $IsExitOnFailure"
    [Security.Principal.WindowsIdentity]$currentProcessToken = [Security.Principal.WindowsIdentity]::GetCurrent()
    [boolean]$isAdmin = [boolean]($currentProcessToken.Groups -contains [Security.Principal.SecurityIdentifier]'S-1-5-32-544')
    Write-HelperOutput -Message "isAdmin: $isAdmin"
    if (!$isAdmin) {
        if ($IsExitOnFailure) {
            Invoke-HelperExit -Message "isAdmin: $isAdmin" -ExitCode 1 -Type 'failure'
        }
    }
    Write-HelperOutput -Message "$functionName execution complete"
    $script:HelperOutputLead = '  '
    return $isAdmin
}

Function Send-HelperMail {
    [CmdletBinding()]
    Param (
        [Parameter(Mandatory = $false)][string[]]$Attachments = @(''),
        [Parameter(Mandatory = $false)][PSCredential]$Credential = $null,
        [Parameter(Mandatory = $false)][string]$Body = 'This is the default helper mail message, and should be replaced with real content.',
        [Parameter(Mandatory = $false)][string]$BCC = '',
        [Parameter(Mandatory = $false)][string]$BodyAsHTML = 'True',
        [Parameter(Mandatory = $false)][string]$CC = '',
        [Parameter(Mandatory = $false)][string]$DeliveryNotificationOption = 'None',
        [Parameter(Mandatory = $false)][string]$EnableSSL = 'True',
        [Parameter(Mandatory = $false)][string]$Encoding = 'utf8',
        [Parameter(Mandatory = $false)][string]$From = '',
        [Parameter(Mandatory = $false)][int]$Port = 25,
        [Parameter(Mandatory = $false)][string]$Priority = 'Normal',
        [Parameter(Mandatory = $false)][string]$Server = 'mail.company.com',
        [Parameter(Mandatory = $false)][string]$Subject = 'Hello World',
        [Parameter(Mandatory = $false)][string]$To = '',
        [Parameter(Mandatory = $false)][boolean]$ExitOnFailure = $true
    )
    [string]$functionName = "{0}" -f $MyInvocation.MyCommand
    $script:HelperOutputLead = '  '
    Write-HelperOutput -Message "$functionName executing..."
    Write-HelperOutput -Message "ExitOnFailure: $ExitOnFailure"
    Write-HelperOutput -Message "Attachments: $Attachments"
    Write-HelperOutput -Message "BCC: $BCC"
    Write-HelperOutput -Message "BodyAsHTML: $BodyAsHTML"
    Write-HelperOutput -Message "CC: $CC"
    Write-HelperOutput -Message "DeliveryNotificationOption: $DeliveryNotificationOption"
    Write-HelperOutput -Message "EnableSSL: $EnableSSL"
    Write-HelperOutput -Message "Encoding: $Encoding"
    Write-HelperOutput -Message "From: $From"
    Write-HelperOutput -Message "Port: $Port"
    Write-HelperOutput -Message "Priority: $Priority"
    Write-HelperOutput -Message "Server: $Server"
    Write-HelperOutput -Message "Subject: $Subject"
    Write-HelperOutput -Message "To: $To"
    if ($null -ne $Credential) {
        Write-HelperOutput -Message "UserName: $($Credential.UserName)"
    }
    [string[]]$toAddresses = @()
    if ($To -ne '') {
        if ($To.Contains(':')) {
            $toAddresses = $To -Split ':'
        }
        else {
            $toAddresses += $To
        }
    }
    if ($toAddresses.Count -eq 0) {
        Write-HelperOutput -Message "Invalid configuration - must supply a valid 'To:' address" -ForegroundColor 'Red'
        if ($ExitOnFailure) {
            Invoke-HelperExit -Code 1
        }
    }
    [string[]]$bccAddresses = @()
    if ($BCC -ne '') {
        if ($BCC.Contains(':')) {
            $bccAddresses = $BCC -Split ':'
        }
        else {
            $bccAddresses += $BCC
        }
    }
    [string[]]$ccAddresses = @()
    if ($CC -ne '') {
        if ($CC.Contains(':')) {
            $ccAddresses = $CC -Split ':'
        }
        else {
            $ccAddresses += $CC
        }
    }
    Write-HelperOutput -Message "Configuring mail parameters"
    $parameters = @{
        From                       = $From
        To                         = $toAddresses
        Subject                    = $Subject
        Body                       = $Body
        DeliveryNotificationOption = $DeliveryNotificationOption
        Encoding                   = $Encoding
        Port                       = $Port
        Priority                   = $Priority
        SmtpServer                 = $Server
    }
    if (($null -ne $bccAddresses) -and ('' -ne $bccAddresses)) {
        $parameters.add("BCC", $bccAddresses)
    }
    if (($null -ne $ccAddresses) -and ('' -ne $ccAddresses)) {
        $parameters.add("CC", $ccAddresses)
    }
    if (($null -ne $Attachments) -and ('' -ne $Attachments)) {
        $parameters.add("Attachments", $Attachments)
    }
    if ($BodyAsHTML) {
        $parameters.add("BodyAsHtml", $null)
    }
    if ($EnableSSL) {
        $parameters.add("UseSSL", $null)
    }
    if ($null -ne $Credential) {
        $parameters.add("Credential", $Credential)
    }
    try {
        Write-HelperOutput -Message "Sending mail"
        # Since we are using splatting, we need to use @ instead of $ in front of the variable name
        Send-MailMessage @parameters
        Write-HelperOutput -Message "Mail sent"
    }
    catch {
        Write-HelperOutput -Message 'An unexpected error occurred while trying to send the mail' -Type 'Warning'
        Write-HelperOutput -Message $_
        if ($ExitOnFailure) {
            Invoke-HelperExit -Code 1
        }
    }
    Write-HelperOutput -Message "   $functionName execution complete"
    $script:HelperOutputLead = ''
}

Function Set-HelperOneTimeScheduledTask {
    [CmdletBinding()]
    Param(
        [Parameter(Mandatory = $true)][string]$Command,
        [Parameter(Mandatory = $false)][string]$CommandArgs = '',
        [Parameter(Mandatory = $false)][PSCredential]$Credential = $null,
        [Parameter(Mandatory = $true)][string]$TaskName,
        [Parameter(Mandatory = $false)][string]$Description = 'One-time self destroying scheduled task',
        [Parameter(Mandatory = $false)][int]$DeleteAfter = 10,
        [Parameter(Mandatory = $false)][int]$DaysToAdd = 1,
        [Parameter(Mandatory = $false)][ValidateSet('Limited', 'Highest')][string]$RunLevel = 'Limited',
        [Parameter(Mandatory = $false)][string]$TimeOfDay = '02:15',
        [Parameter(Mandatory = $false)][boolean]$ExitOnFailure = $true
    )
    [string]$functionName = "{0}" -f $MyInvocation.MyCommand
    $script:HelperOutputLead = '  '
    Write-HelperOutput -Message "$functionName executing..."
    Write-HelperOutput -Message "Command: $Command"
    Write-HelperOutput -Message "CommandArgs: $CommandArgs"
    Write-HelperOutput -Message "TaskName: $TaskName"
    Write-HelperOutput -Message "Description: $Description"
    Write-HelperOutput -Message "DeleteAfter: $DeleteAfter"
    Write-HelperOutput -Message "DaysToAdd: $DaysToAdd"
    Write-HelperOutput -Message "RunLevel: $RunLevel"
    Write-HelperOutput -Message "TimeOfDay: $TimeOfDay"
    Write-HelperOutput -Message "ExitOnFailure: $ExitOnFailure"
    [System.DateTime]$runDate = (Get-Date).AddDays($DaysToAdd)
    Write-HelperOutput -Message "runDate: $runDate"
    [string]$dateElement = Get-Date -Date $runDate -Format "yyyy-MM-dd"
    Write-HelperOutput -Message "dateElement: $dateElement"
    [System.DateTime]$triggerAt = [System.DateTime]"$($dateElement)T$($TimeOfDay)"
    Write-HelperOutput -Message "triggerAt: $triggerAt"
    [string]$expiresAt = $triggerAt.AddMinutes($DeleteAfter).ToString('s')
    Write-HelperOutput -Message "expiresAt: $expiresAt"
    [string]$name = "$TaskName $dateElement"
    [string]$taskUser = "NT AUTHORITY\SYSTEM"
    [string]$taskUserPassword = ''
    if ($null -ne $Credential) {
        $taskUser = "$($Credential.UserName)"
        $taskUserPassword = "$($Credential.GetNetworkCredential().password)"
    }
    Write-HelperOutput -Message "taskUser: $taskUser"
    [PSCustomObject]$taskObject = Get-ScheduledTask -TaskName "$name" -TaskPath "\" -ErrorAction 'SilentlyContinue' | Select-Object *
    if ($null -ne $taskObject) {
        Write-HelperOutput -Message "An existing '$name' scheduled task already exists" -Type 'Error'
        $taskUserPassword = $null
        if ($ExitOnFailure) {
            Invoke-HelperExit -Code 1
        }
    }
    Write-HelperOutput -Message "Creating '$name' scheduled task"
    try {
        $null = Register-ScheduledTask -TaskName "$name" -InputObject (
            (
                New-ScheduledTask -Description "$Description" -Action (
                    New-ScheduledTaskAction -Execute "$Command" -Argument "$CommandArgs"
                ) -Trigger (
                    New-ScheduledTaskTrigger -Once -At $triggerAt
                ) -Principal (
                    New-ScheduledTaskPrincipal -UserId "$taskUser" -RunLevel $RunLevel
                ) -Settings (
                    New-ScheduledTaskSettingsSet -DeleteExpiredTaskAfter 00:00:03 # Delete the task three seconds after trigger expires
                ) 
            ) | ForEach-Object { $_.Triggers[0].EndBoundary = "$expiresAt"; $_ }
        )
        $taskObject = Get-ScheduledTask -TaskName "$name" -TaskPath "\" -ErrorAction 'SilentlyContinue' | Select-Object *
        if ($null -ne $taskObject) {
            Write-HelperOutput -Message "Task created"
        }
        else {
            Write-HelperOutput -Message "Failed to create the '$name' scheduled task" -Type 'Error'
            $taskUserPassword = $null
            if ($ExitOnFailure) {
                Invoke-HelperExit -Code 1
            }
        }
    }
    catch {
        Write-HelperOutput -Message "An unexpected error occured creating the '$name' scheduled task" -Type 'Error'
        $taskUserPassword = $null
        if ($ExitOnFailure) {
            Invoke-HelperExit -Code 1
        }
    }
    if ($null -ne $taskObject) {
        try {
            Write-HelperOutput -Message "Updating the scheduled task"
            if ($taskUserPassword -eq '') {
                $null = Set-ScheduledTask -TaskName "$name" -User "$taskUser"
            }
            else {
                $null = Set-ScheduledTask -TaskName "$name" -User "$taskUser" -Password "$taskUserPassword"
            }
            Write-HelperOutput -Message "Successfully set the scheduled task" -ForegroundColor 'Green'
        }
        catch {
            Write-HelperOutput -Message "An unexpected error occured updating the existing '$name' scheduled task" -Type 'Error'
            $taskUserPassword = $null
            if ($ExitOnFailure) {
                Invoke-HelperExit -Code 1
            }
        }
    }
    $taskUser = $null
    $taskUserPassword = $null
    Write-HelperOutput "$functionName execution complete"
    $script:HelperOutputLead = ''
}

Function Initialize-Helper {
    [CmdletBinding()]
    Param
    (
        [Parameter(Mandatory = $true)][string]$Identifier,
        [Parameter(Mandatory = $false)][boolean]$LogOutput = $true,
        [Parameter(Mandatory = $false)][string]$LogFolder = "$($env:WinDir)\Temp"
    )
    [string]$functionName = "{0}" -f $MyInvocation.MyCommand
    $script:HelperOutputLead = '  '
    Initialize-HelperOutput -Identifier "$Identifier" -LogOutput $LogOutput -LogFolder $LogFolder
    Write-HelperOutput -Message "$functionName executing..."
    Write-HelperOutput -Message "Identifier: $Identifier"
    Write-HelperOutput -Message "LogOutput: $LogOutput"
    Write-HelperOutput -Message "LogFolder: $LogFolder"
    try {
        Write-HelperOutput -Message "Operating System: $((Get-CimInstance -ClassName CIM_OperatingSystem).Caption)"
        Write-HelperOutput -Message "OS Service Pack: $((Get-CimInstance -ClassName CIM_OperatingSystem).ServicePackMajorVersion)"
        Write-HelperOutput -Message "PowerShell Version: $($PSVersionTable.PSVersion)"
        Write-HelperOutput -Message "PowerShell Edition: $($PSVersionTable.PSEdition)"
        Write-HelperOutput -Message "Computer Name: $($env:ComputerName)"
        Write-HelperOutput -Message "User Domain: $($env:UserDomain)"
        [boolean]$isAdmin = Test-HelperAdmin -IsExitOnFailure $false
        if (!$isAdmin) {
            Write-HelperOutput -Message "If appropriate, run the console session with admin rights ('Run as administrator')" -Type 'Console' -ForegroundColor 'DarkCyan'
        }
    }
    catch {
        Write-HelperOutput -Message "$_"
        Invoke-HelperExit -Message "An unexpected error occured while initializing" -Type 'failure'
    }
    Write-HelperOutput -Message "$functionName execution complete"
}