# @summary Shadow Copy
#
# Defines a Windows drive shadow copy configuration
#
# @param auto_retry_minutes
#   The maximum amount of time (in minutes) to attempt to create the shadow copy
#   Example: 25
#
# @param ensure
#   denotes the desired state of shadow copy for the supplied drive letter
#   Example: 'present'
#
# @param letter
#   Enumerated string containing the letter of the drive to configure
#   Example: 'E'
#
# @param max_size
#   Denotes the maximum amount of storage space to consume
#      If limited, must be 300 MB or greater
#      Do not set if wanting Unlimited
#   Example: 450
#
# @param max_size_type
#   Denotes the type of max_size value supplied
#   Example: 'GB'
#
# @param on
#   Enumerated string containing the letter of the drive to store the shadow copy
#   Example: 'K'
#
# @param task_compatibility
#   String containing the compatibiltiy type for the associated scheduled task
#   Example: 'Win8'
#
# @param task_execution_time_limit
#   String of the amount of time that the associated scheduled task is allowed for completion
#      D=Days, H=Hours
#   Example: '4H'
#
# @param task_priority
#   The priority level given to the asssociated scheduled task
#   Example: 5
#
# @param triggers
#   Array of hashes containing the scheduled task trigger configurations
#   Example: [
#     {
#       at: '11:45pm',
#       days_of_week: ['Friday'],
#       weeks_interval: 3,
#     },
#     {
#       at: '8:15am',
#       days_of_week: ['Monday', 'Wednesday'],
#       weeks_interval: 1,
#     },
#   ],
#
define windows_drives::defines::shadow_copy (
  Enum['A', 'B', 'C', 'D', 'E', 'F',
    'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N',
    'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V',
  'W', 'X', 'Y', 'Z'] $letter,
  Integer $auto_retry_minutes = 15,
  Enum['absent', 'present', 'unmanaged'] $ensure = 'unmanaged',
  Integer $max_size = undef,
  Enum['bytes', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB'] $max_size_type = 'MB',
  Enum['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J',
    'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U',
  'V', 'W', 'X', 'Y', 'Z'] $on = $letter,
  Enum['At', 'V1', 'Vista', 'Win7', 'Win8'] $task_compatibility = 'V1',
  Enum['1H', '2H', '4H', '8H', '12H', '1D', '3D'] $task_execution_time_limit = '8H',
  Integer[0,10] $task_priority = 7,
  Array[Hash] $triggers = [],
) {
  if ($windows_drives::log_output) {
    notify { "${module_name} - shadow_copy - letter: '${letter}'": }
    notify { "${module_name} - shadow_copy - auto_retry_minutes: '${auto_retry_minutes}'": }
    notify { "${module_name} - shadow_copy - ensure: '${ensure}'": }
    notify { "${module_name} - shadow_copy - max_size: '${max_size}'": }
    notify { "${module_name} - shadow_copy - max_size_type: '${max_size_type}'": }
    notify { "${module_name} - shadow_copy - on: '${on}'": }
    notify { "${module_name} - shadow_copy - task_compatibility: '${task_compatibility}'": }
    notify { "${module_name} - shadow_copy - task_execution_time_limit: '${task_execution_time_limit}'": }
    notify { "${module_name} - shadow_copy - task_priority: '${task_priority}'": }
    notify { "${module_name} - shadow_copy - number of triggers: '${triggers.length}'": }
  }
  if (($task_execution_time_limit !~ /D$/) and ($task_execution_time_limit !~ /H$/)) {
    fail "Invalid 'execution_time_limit' (${task_execution_time_limit}) supplied for '${title}' (must end with valid time designator: D,H)"
  }
  # validate supplied trigger data
  $triggers.each | Hash $trigger_hash | {
    windows_drives::defines::trigger { "${letter} at ${trigger_hash['at']}":
      * => $trigger_hash,
    }
  }
  $epp_param_hash = {
    'auto_retry_minutes'        => $auto_retry_minutes,
    'ensure'                    => $ensure,
    'identifier'                => $module_name,
    'letter'                    => $letter,
    'log_output'                => $windows_drives::log_output,
    'max_size'                  => $max_size,
    'max_size_type'             => $max_size_type,
    'on'                        => $on,
    'task_compatibility'        => $task_compatibility,
    'task_execution_time_limit' => $task_execution_time_limit,
    'task_priority'             => $task_priority,
    'triggers'                  => $triggers,
  }
  exec { "Set shadow copy for ${letter}":
    provider  => powershell,
    command   => Sensitive("${file("${module_name}/helpers.ps1")};
                  ${epp("${module_name}/shadow_copy_command.ps1.epp", $epp_param_hash)}"
    ),
    onlyif    => Sensitive("${file("${module_name}/helpers.ps1")};
                  ${epp("${module_name}/shadow_copy_onlyif.ps1.epp", $epp_param_hash)}"
    ),
    logoutput => $windows_drives::log_output,
  }
}
